@@ -1,20 +1,20 @@
 /* assuan-buffer.c - read and send data
- * Copyright (C) 2001, 2002, 2003, 2004, 2006 Free Software Foundation, Inc.
- *
- * This file is part of Assuan.
- *
- * Assuan is free software; you can redistribute it and/or modify it
- * under the terms of the GNU Lesser General Public License as
- * published by the Free Software Foundation; either version 2.1 of
- * the License, or (at your option) any later version.
- *
- * Assuan is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this program; if not, see <http://www.gnu.org/licenses/>.
+   Copyright (C) 2001-2004, 2006, 2009 Free Software Foundation, Inc.
+
+   This file is part of Assuan.
+
+   Assuan is free software; you can redistribute it and/or modify it
+   under the terms of the GNU Lesser General Public License as
+   published by the Free Software Foundation; either version 2.1 of
+   the License, or (at your option) any later version.
+
+   Assuan is distributed in the hope that it will be useful, but
+   WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+   Lesser General Public License for more details.
+
+   You should have received a copy of the GNU Lesser General Public
+   License along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include <config.h>
@@ -31,13 +31,17 @@
 
 
 /* Extended version of write(2) to guarantee that all bytes are
-   written.  Returns 0 on success or -1 and ERRNO on failure. */
+   written.  Returns 0 on success or -1 and ERRNO on failure.  NOTE:
+   This function does not return the number of bytes written, so any
+   error must be treated as fatal for this connection as the state of
+   the receiver is unknown.  This works best if blocking is allowed
+   (so EAGAIN cannot occur).  */
 static int
 writen (assuan_context_t ctx, const char *buffer, size_t length)
 {
   while (length)
     {
-      ssize_t nwritten = ctx->io->writefnc (ctx, buffer, length);
+      ssize_t nwritten = ctx->engine.writefnc (ctx, buffer, length);
       
       if (nwritten < 0)
         {
@@ -66,7 +70,7 @@ readline (assuan_context_t ctx, char *buf, size_t buflen,
   *r_nread = 0;
   while (nleft > 0)
     {
-      ssize_t n = ctx->io->readfnc (ctx, buf, nleft);
+      ssize_t n = ctx->engine.readfnc (ctx, buf, nleft);
 
       if (n < 0)
         {
@@ -79,6 +83,7 @@ readline (assuan_context_t ctx, char *buf, size_t buflen,
           *r_eof = 1;
           break; /* allow incomplete lines */
         }
+
       p = buf;
       nleft -= n;
       buf += n;
@@ -88,22 +93,22 @@ readline (assuan_context_t ctx, char *buf, size_t buflen,
       if (p)
         break; /* at least one full line available - that's enough for now */
     }
-
   return 0;
 }
 
 
-/* Function returns an Assuan error.  */
-assuan_error_t
+/* Read a line with buffering of partial lines.  Function returns an
+   Assuan error.  */
+gpg_error_t
 _assuan_read_line (assuan_context_t ctx)
 {
+  gpg_error_t rc = 0;
   char *line = ctx->inbound.line;
   int nread, atticlen;
-  int rc;
   char *endp = 0;
 
   if (ctx->inbound.eof)
-    return _assuan_error (-1);
+    return _assuan_error (ctx, GPG_ERR_EOF);
 
   atticlen = ctx->inbound.attic.linelen;
   if (atticlen)
@@ -113,15 +118,14 @@ _assuan_read_line (assuan_context_t ctx)
 
       endp = memchr (line, '\n', atticlen);
       if (endp)
-	/* Found another line in the attic.  */
 	{
-	  rc = 0;
+	  /* Found another line in the attic.  */
 	  nread = atticlen;
 	  atticlen = 0;
 	}
       else
-	/* There is pending data but not a full line.  */
         {
+	  /* There is pending data but not a full line.  */
           assert (atticlen < LINELENGTH);
           rc = readline (ctx, line + atticlen,
 			 LINELENGTH - atticlen, &nread, &ctx->inbound.eof);
@@ -143,14 +147,16 @@ _assuan_read_line (assuan_context_t ctx)
 
       if (saved_errno == EAGAIN)
         {
-          /* We have to save a partial line.  */
+          /* We have to save a partial line.  Due to readline's
+	     behaviour, we know that this is not a complete line yet
+	     (no newline).  So we don't set PENDING to true.  */
           memcpy (ctx->inbound.attic.line, line, atticlen + nread);
           ctx->inbound.attic.pending = 0;
           ctx->inbound.attic.linelen = atticlen + nread;
         }
 
       errno = saved_errno;
-      return _assuan_error (ASSUAN_Read_Error);
+      return _assuan_error (ctx, gpg_err_code_from_syserror ());
     }
   if (!nread)
     {
@@ -159,7 +165,8 @@ _assuan_read_line (assuan_context_t ctx)
 	fprintf (ctx->log_fp, "%s[%u.%d] DBG: <- [EOF]\n",
 		 assuan_get_assuan_log_prefix (),
                  (unsigned int)getpid (), (int)ctx->inbound.fd);
-      return _assuan_error (-1);
+
+      return _assuan_error (ctx, GPG_ERR_EOF);
     }
 
   ctx->inbound.attic.pending = 0;
@@ -190,20 +197,20 @@ _assuan_read_line (assuan_context_t ctx)
 
       ctx->inbound.linelen = endp - line;
 
-      monitor_result = (ctx->io_monitor
-                        ? ctx->io_monitor (ctx, 0,
-                                           ctx->inbound.line,
-                                           ctx->inbound.linelen)
-                        : 0);
-      if ( (monitor_result & 2) )
+      monitor_result = 0;
+      if (ctx->io_monitor)
+	monitor_result = ctx->io_monitor (ctx, ctx->io_monitor_data, 0,
+					  ctx->inbound.line,
+					  ctx->inbound.linelen);
+      if (monitor_result & ASSUAN_IO_MONITOR_IGNORE)
         ctx->inbound.linelen = 0;
       
-      if (ctx->log_fp && !(monitor_result & 1))
+      if (ctx->log_fp && !(monitor_result & ASSUAN_IO_MONITOR_NOLOG))
 	{
 	  fprintf (ctx->log_fp, "%s[%u.%d] DBG: <- ",
 		   assuan_get_assuan_log_prefix (),
                    (unsigned int)getpid (), (int)ctx->inbound.fd);
-	  if (ctx->confidential)
+	  if (ctx->flags.confidential)
 	    fputs ("[Confidential data not shown]", ctx->log_fp);
 	  else
 	    _assuan_log_print_buffer (ctx->log_fp,
@@ -221,9 +228,9 @@ _assuan_read_line (assuan_context_t ctx)
                  (unsigned int)getpid (), (int)ctx->inbound.fd);
       *line = 0;
       ctx->inbound.linelen = 0;
-      return _assuan_error (ctx->inbound.eof 
-                            ? ASSUAN_Line_Not_Terminated
-                            : ASSUAN_Line_Too_Long);
+      return _assuan_error (ctx, ctx->inbound.eof 
+			    ? GPG_ERR_ASS_INCOMPLETE_LINE
+			    : GPG_ERR_ASS_LINE_TOO_LONG);
     }
 }
 
@@ -237,22 +244,23 @@ _assuan_read_line (assuan_context_t ctx)
    Returns 0 on success or an assuan error code.
    See also: assuan_pending_line().
 */
-assuan_error_t
+gpg_error_t
 assuan_read_line (assuan_context_t ctx, char **line, size_t *linelen)
 {
-  assuan_error_t err;
+  gpg_error_t err;
 
   if (!ctx)
-    return _assuan_error (ASSUAN_Invalid_Value);
+    return _assuan_error (ctx, GPG_ERR_ASS_INV_VALUE);
 
   do
     {
       err = _assuan_read_line (ctx);
     }
-  while (_assuan_error_is_eagain (err));
+  while (_assuan_error_is_eagain (ctx, err));
 
   *line = ctx->inbound.line;
   *linelen = ctx->inbound.linelen;
+
   return err;
 }
 
@@ -266,11 +274,11 @@ assuan_pending_line (assuan_context_t ctx)
 }
 
 
-assuan_error_t 
+gpg_error_t 
 _assuan_write_line (assuan_context_t ctx, const char *prefix,
                     const char *line, size_t len)
 {
-  assuan_error_t rc = 0;
+  gpg_error_t rc = 0;
   size_t prefixlen = prefix? strlen (prefix):0;
   unsigned int monitor_result;
 
@@ -288,17 +296,17 @@ _assuan_write_line (assuan_context_t ctx, const char *prefix,
         len = ASSUAN_LINELENGTH - prefixlen - 2 - 1;
     }
 
-  monitor_result = (ctx->io_monitor
-                    ? ctx->io_monitor (ctx, 1, line, len)
-                    : 0);
+  monitor_result = 0;
+  if (ctx->io_monitor)
+    monitor_result = ctx->io_monitor (ctx, ctx->io_monitor_data, 1, line, len);
 
   /* Fixme: we should do some kind of line buffering.  */
-  if (ctx->log_fp && !(monitor_result & 1))
+  if (ctx->log_fp && !(monitor_result & ASSUAN_IO_MONITOR_NOLOG))
     {
       fprintf (ctx->log_fp, "%s[%u.%d] DBG: -> ",
 	       assuan_get_assuan_log_prefix (),
                (unsigned int)getpid (), (int)ctx->inbound.fd);
-      if (ctx->confidential)
+      if (ctx->flags.confidential)
 	fputs ("[Confidential data not shown]", ctx->log_fp);
       else
         {
@@ -309,47 +317,47 @@ _assuan_write_line (assuan_context_t ctx, const char *prefix,
       putc ('\n', ctx->log_fp);
     }
 
-  if (prefixlen && !(monitor_result & 2))
+  if (prefixlen && !(monitor_result & ASSUAN_IO_MONITOR_IGNORE))
     {
       rc = writen (ctx, prefix, prefixlen);
       if (rc)
-        rc = _assuan_error (ASSUAN_Write_Error);
+	rc = _assuan_error (ctx, gpg_err_code_from_syserror ());
     }
-  if (!rc && !(monitor_result & 2))
+  if (!rc && !(monitor_result & ASSUAN_IO_MONITOR_IGNORE))
     {
       rc = writen (ctx, line, len);
       if (rc)
-        rc = _assuan_error (ASSUAN_Write_Error);
+	rc = _assuan_error (ctx, gpg_err_code_from_syserror ());
       if (!rc)
         {
           rc = writen (ctx, "\n", 1);
           if (rc)
-            rc = _assuan_error (ASSUAN_Write_Error);
+	    rc = _assuan_error (ctx, gpg_err_code_from_syserror ());
         }
     }
   return rc;
 }
 
 
-assuan_error_t 
+gpg_error_t 
 assuan_write_line (assuan_context_t ctx, const char *line)
 {
   size_t len;
-  const char *s;
+  const char *str;
 
-  if (!ctx)
-    return _assuan_error (ASSUAN_Invalid_Value);
+  if (! ctx)
+    return _assuan_error (ctx, GPG_ERR_ASS_INV_VALUE);
 
   /* Make sure that we never take a LF from the user - this might
      violate the protocol. */
-  s = strchr (line, '\n');
-  len = s? (s-line) : strlen (line);
+  str = strchr (line, '\n');
+  len = str ? (str - line) : strlen (line);
 
-  if (ctx->log_fp && s)
+  if (ctx->log_fp && str)
     fprintf (ctx->log_fp, "%s[%u.%d] DBG: -> "
              "[supplied line contained a LF - truncated]\n",
              assuan_get_assuan_log_prefix (),
-             (unsigned int)getpid (), (int)ctx->inbound.fd);
+             (unsigned int) getpid (), (int) ctx->inbound.fd);
 
   return _assuan_write_line (ctx, NULL, line, len);
 }
@@ -403,20 +411,20 @@ _assuan_cookie_write_data (void *cookie, const char *buffer, size_t orig_size)
         }
       
       
-      monitor_result = (ctx->io_monitor
-                        ? ctx->io_monitor (ctx, 1,
-                                           ctx->outbound.data.line, linelen)
-                        : 0);
+      monitor_result = 0;
+      if (ctx->io_monitor)
+	monitor_result = ctx->io_monitor (ctx, ctx->io_monitor_data, 1,
+					  ctx->outbound.data.line, linelen);
 
       if (linelen >= LINELENGTH-2-2)
         {
-          if (ctx->log_fp && !(monitor_result & 1))
+          if (ctx->log_fp && !(monitor_result & ASSUAN_IO_MONITOR_NOLOG))
             {
 	      fprintf (ctx->log_fp, "%s[%u.%d] DBG: -> ",
 		       assuan_get_assuan_log_prefix (),
                        (unsigned int)getpid (), (int)ctx->inbound.fd);
 
-              if (ctx->confidential)
+              if (ctx->flags.confidential)
                 fputs ("[Confidential data not shown]", ctx->log_fp);
               else 
                 _assuan_log_print_buffer (ctx->log_fp, 
@@ -426,10 +434,10 @@ _assuan_cookie_write_data (void *cookie, const char *buffer, size_t orig_size)
             }
           *line++ = '\n';
           linelen++;
-          if ( !(monitor_result & 2)
+          if ( !(monitor_result & ASSUAN_IO_MONITOR_IGNORE)
                && writen (ctx, ctx->outbound.data.line, linelen))
             {
-              ctx->outbound.data.error = _assuan_error (ASSUAN_Write_Error);
+              ctx->outbound.data.error = gpg_err_code_from_syserror ();
               return 0;
             }
           line = ctx->outbound.data.line;
@@ -438,7 +446,7 @@ _assuan_cookie_write_data (void *cookie, const char *buffer, size_t orig_size)
     }
 
   ctx->outbound.data.linelen = linelen;
-  return (int)orig_size;
+  return (int) orig_size;
 }
 
 
@@ -459,19 +467,19 @@ _assuan_cookie_write_flush (void *cookie)
   linelen = ctx->outbound.data.linelen;
   line += linelen;
 
-  monitor_result = (ctx->io_monitor
-                    ? ctx->io_monitor (ctx, 1,
-                                       ctx->outbound.data.line, linelen)
-                    : 0);
+  monitor_result = 0;
+  if (ctx->io_monitor)
+    monitor_result = ctx->io_monitor (ctx, ctx->io_monitor_data, 1,
+				      ctx->outbound.data.line, linelen);
   
   if (linelen)
     {
-      if (ctx->log_fp && !(monitor_result & 1))
+      if (ctx->log_fp && !(monitor_result & ASSUAN_IO_MONITOR_NOLOG))
 	{
 	  fprintf (ctx->log_fp, "%s[%u.%d] DBG: -> ",
 		   assuan_get_assuan_log_prefix (),
                    (unsigned int)getpid (), (int)ctx->inbound.fd);
-	  if (ctx->confidential)
+	  if (ctx->flags.confidential)
 	    fputs ("[Confidential data not shown]", ctx->log_fp);
 	  else
 	    _assuan_log_print_buffer (ctx->log_fp,
@@ -480,10 +488,10 @@ _assuan_cookie_write_flush (void *cookie)
 	}
       *line++ = '\n';
       linelen++;
-      if ( !(monitor_result & 2)
+      if (! (monitor_result & ASSUAN_IO_MONITOR_IGNORE)
            && writen (ctx, ctx->outbound.data.line, linelen))
         {
-          ctx->outbound.data.error = _assuan_error (ASSUAN_Write_Error);
+          ctx->outbound.data.error = gpg_err_code_from_syserror ();
           return 0;
         }
       ctx->outbound.data.linelen = 0;
@@ -507,16 +515,19 @@ _assuan_cookie_write_flush (void *cookie)
  * a INQUIRE response.  However, when assuan_transact() is used, this
  * function takes care of sending END itself.
  * 
+ * If BUFFER is NULL and LENGTH is 1 and we are a client, a "CAN" is
+ * send instead of an "END".
+ * 
  * Return value: 0 on success or an error code
  **/
 
-assuan_error_t
+gpg_error_t
 assuan_send_data (assuan_context_t ctx, const void *buffer, size_t length)
 {
   if (!ctx)
-    return _assuan_error (ASSUAN_Invalid_Value);
-  if (!buffer && length)
-    return _assuan_error (ASSUAN_Invalid_Value);
+    return _assuan_error (ctx, GPG_ERR_ASS_INV_VALUE);
+  if (!buffer && length > 1)
+    return _assuan_error (ctx, GPG_ERR_ASS_INV_VALUE);
 
   if (!buffer)
     { /* flush what we have */
@@ -524,7 +535,7 @@ assuan_send_data (assuan_context_t ctx, const void *buffer, size_t length)
       if (ctx->outbound.data.error)
         return ctx->outbound.data.error;
       if (!ctx->is_server)
-        return assuan_write_line (ctx, "END");
+        return assuan_write_line (ctx, length == 1? "CAN":"END");
     }
   else
     {
@@ -536,7 +547,7 @@ assuan_send_data (assuan_context_t ctx, const void *buffer, size_t length)
   return 0;
 }
 
-assuan_error_t
+gpg_error_t
 assuan_sendfd (assuan_context_t ctx, assuan_fd_t fd)
 {
   /* It is explicitly allowed to use (NULL, -1) as a runtime test to
@@ -545,22 +556,22 @@ assuan_sendfd (assuan_context_t ctx, assuan_fd_t fd)
 #ifdef USE_DESCRIPTOR_PASSING
     return 0;
 #else
-    return _assuan_error (ASSUAN_Not_Implemented);
+  return _assuan_error (ctx, GPG_ERR_NOT_IMPLEMENTED);
 #endif
 
-  if (! ctx->io->sendfd)
-    return set_error (ctx, Not_Implemented,
+  if (! ctx->engine.sendfd)
+    return set_error (ctx, GPG_ERR_NOT_IMPLEMENTED,
 		      "server does not support sending and receiving "
 		      "of file descriptors");
-  return ctx->io->sendfd (ctx, fd);
+  return ctx->engine.sendfd (ctx, fd);
 }
 
-assuan_error_t
+gpg_error_t
 assuan_receivefd (assuan_context_t ctx, assuan_fd_t *fd)
 {
-  if (! ctx->io->receivefd)
-    return set_error (ctx, Not_Implemented,
+  if (! ctx->engine.receivefd)
+    return set_error (ctx, GPG_ERR_NOT_IMPLEMENTED,
 		      "server does not support sending and receiving "
 		      "of file descriptors");
-  return ctx->io->receivefd (ctx, fd);
+  return ctx->engine.receivefd (ctx, fd);
 }
