@@ -1,23 +1,26 @@
 /* assuan-socket-server.c - Assuan socket based server
- *	Copyright (C) 2002, 2007 Free Software Foundation, Inc.
- *
- * This file is part of Assuan.
- *
- * Assuan is free software; you can redistribute it and/or modify it
- * under the terms of the GNU Lesser General Public License as
- * published by the Free Software Foundation; either version 2.1 of
- * the License, or (at your option) any later version.
- *
- * Assuan is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this program; if not, see <http://www.gnu.org/licenses/>.
+   Copyright (C) 2002, 2007, 2009 Free Software Foundation, Inc.
+
+   This file is part of Assuan.
+
+   Assuan is free software; you can redistribute it and/or modify it
+   under the terms of the GNU Lesser General Public License as
+   published by the Free Software Foundation; either version 2.1 of
+   the License, or (at your option) any later version.
+
+   Assuan is distributed in the hope that it will be useful, but
+   WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+   Lesser General Public License for more details.
+
+   You should have received a copy of the GNU Lesser General Public
+   License along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
+#ifdef HAVE_CONFIG_H
 #include <config.h>
+#endif
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <errno.h>
@@ -35,19 +38,15 @@
 # include <sys/un.h>
 #endif
 
-
+#include "debug.h"
 #include "assuan-defs.h"
 
-static struct assuan_io io = { _assuan_simple_read,
-			       _assuan_simple_write };
-
-
-static int
+static gpg_error_t
 accept_connection_bottom (assuan_context_t ctx)
 {
   assuan_fd_t fd = ctx->connected_fd;
 
-  ctx->peercred.valid = 0;
+  ctx->peercred_valid = 0;
 #ifdef HAVE_SO_PEERCRED
   {
     struct ucred cr; 
@@ -58,11 +57,11 @@ accept_connection_bottom (assuan_context_t ctx)
          ctx->peercred.pid = cr.pid;
          ctx->peercred.uid = cr.uid;
          ctx->peercred.gid = cr.gid;
-         ctx->peercred.valid = 1;
+         ctx->peercred_valid = 1;
 
          /* This overrides any already set PID if the function returns
             a valid one. */
-         if (cr.pid != (pid_t)-1 && cr.pid) 
+         if (cr.pid != ASSUAN_INVALID_PID && cr.pid) 
            ctx->pid = cr.pid;
       }
   }
@@ -78,13 +77,13 @@ accept_connection_bottom (assuan_context_t ctx)
   ctx->outbound.data.linelen = 0;
   ctx->outbound.data.error = 0;
   
-  ctx->confidential = 0;
+  ctx->flags.confidential = 0;
 
   return 0;
 }
 
 
-static int
+static gpg_error_t
 accept_connection (assuan_context_t ctx)
 {
   assuan_fd_t fd;
@@ -95,81 +94,53 @@ accept_connection (assuan_context_t ctx)
                              (struct sockaddr*)&clnt_addr, &len ));
   if (fd == ASSUAN_INVALID_FD)
     {
-      ctx->os_errno = errno;
-      return _assuan_error (ASSUAN_Accept_Failed);
+      return _assuan_error (ctx, gpg_err_code_from_syserror ());
     }
-  if (_assuan_sock_check_nonce (fd, &ctx->listen_nonce))
+  if (_assuan_sock_check_nonce (ctx, fd, &ctx->listen_nonce))
     {
-      _assuan_close (fd);
-      ctx->os_errno = EACCES;
-      return _assuan_error (ASSUAN_Accept_Failed);
+      _assuan_close (ctx, fd);
+      return _assuan_error (ctx, GPG_ERR_ASS_ACCEPT_FAILED);
     }
 
   ctx->connected_fd = fd;
   return accept_connection_bottom (ctx);
 }
 
-static int
-finish_connection (assuan_context_t ctx)
-{
-  if (ctx->inbound.fd != ASSUAN_INVALID_FD)
-    {
-      _assuan_close (ctx->inbound.fd);
-    }
-  ctx->inbound.fd = ASSUAN_INVALID_FD;
-  ctx->outbound.fd = ASSUAN_INVALID_FD;
-  return 0;
-}
-
-
-static void
-deinit_socket_server (assuan_context_t ctx)
-{
-  finish_connection (ctx);
-}
-
-/* Initialize a server for the socket LISTEN_FD which has already be
-   put into listen mode */
-int
-assuan_init_socket_server (assuan_context_t *r_ctx, assuan_fd_t listen_fd)
-{
-  return assuan_init_socket_server_ext (r_ctx, listen_fd, 0);
-}
-
-/* Initialize a server using the already accepted socket FD.  This
-   function is deprecated. */
-int
-assuan_init_connected_socket_server (assuan_context_t *r_ctx, assuan_fd_t fd)
-{
-  return assuan_init_socket_server_ext (r_ctx, fd, 2);
-}
-
 
 /* 
    Flag bits: 0 - use sendmsg/recvmsg to allow descriptor passing
               1 - FD has already been accepted.
 */
-int
-assuan_init_socket_server_ext (assuan_context_t *r_ctx, assuan_fd_t fd,
-                               unsigned int flags)
+gpg_error_t
+assuan_init_socket_server (assuan_context_t ctx, assuan_fd_t fd,
+			   unsigned int flags)
 {
-  assuan_context_t ctx;
-  int rc;
+  gpg_error_t rc;
+  TRACE_BEG2 (ctx, ASSUAN_LOG_CTX, "assuan_init_socket_server", ctx,
+	      "fd=0x%x, flags=0x%x", fd, flags);
+  
+  rc = _assuan_register_std_commands (ctx);
+  if (rc)
+    return TRACE_ERR (rc);
 
-  *r_ctx = NULL;
-  ctx = xtrycalloc (1, sizeof *ctx);
-  if (!ctx)
-    return _assuan_error (ASSUAN_Out_Of_Core);
+  ctx->engine.release = _assuan_server_release;
+  ctx->engine.readfnc = _assuan_simple_read;
+  ctx->engine.writefnc = _assuan_simple_write;
+  ctx->engine.sendfd = NULL;
+  ctx->engine.receivefd = NULL;
   ctx->is_server = 1;
-  if ((flags & 2))
-    ctx->pipe_mode = 1; /* We want a second accept to indicate EOF. */
+  if (flags & ASSUAN_SOCKET_SERVER_ACCEPTED)
+    /* We want a second accept to indicate EOF. */
+    ctx->max_accepts = 1;
+  else
+    ctx->max_accepts = -1;
   ctx->input_fd = ASSUAN_INVALID_FD;
   ctx->output_fd = ASSUAN_INVALID_FD;
 
   ctx->inbound.fd = ASSUAN_INVALID_FD;
   ctx->outbound.fd = ASSUAN_INVALID_FD;
 
-  if ((flags & 2))
+  if (flags & ASSUAN_SOCKET_SERVER_ACCEPTED)
     {
       ctx->listen_fd = ASSUAN_INVALID_FD;
       ctx->connected_fd = fd;
@@ -179,22 +150,18 @@ assuan_init_socket_server_ext (assuan_context_t *r_ctx, assuan_fd_t fd,
       ctx->listen_fd = fd;
       ctx->connected_fd = ASSUAN_INVALID_FD;
     }
-  ctx->deinit_handler = (flags & 1)? _assuan_uds_deinit:deinit_socket_server;
-  ctx->accept_handler = ((flags & 2)
+  ctx->accept_handler = ((flags & ASSUAN_SOCKET_SERVER_ACCEPTED)
                          ? accept_connection_bottom 
                          : accept_connection);
-  ctx->finish_handler = finish_connection;
+  ctx->finish_handler = _assuan_server_finish;
 
-  ctx->io = &io;
-  if ((flags & 1))
+  if (flags & ASSUAN_SOCKET_SERVER_FDPASSING)
     _assuan_init_uds_io (ctx);
 
   rc = _assuan_register_std_commands (ctx);
   if (rc)
-    xfree (ctx);
-  else
-    *r_ctx = ctx;
-  return rc;
+    _assuan_reset (ctx);
+  return TRACE_ERR (rc);
 }
 
 
